import textwrap

import pytest

from conan.errors import ConanException
from conan.internal.model.options import Options
from conan.api.model import RecipeReference


class TestOptions:

    @pytest.fixture(autouse=True)
    def _setup(self):
        options = {"static": [True, False], "optimized": [2, 3, 4], "path": ["ANY"]}
        values = {"static": True, "optimized": 3, "path": "mypath"}
        self.sut = Options(options, values)

    def test_booleans(self):
        assert self.sut.static
        assert self.sut.static == True
        assert self.sut.static != False
        assert bool(self.sut.static)

        assert self.sut.optimized
        assert self.sut.optimized == 3
        assert self.sut.optimized != 2

        assert self.sut.path == "mypath"
        assert self.sut.path != "otherpath"

    def test_invalid_value(self):
        with pytest.raises(ConanException) as e:
            assert self.sut.static != 1
        assert "'1' is not a valid 'options.static' value" in str(e.value)

        with pytest.raises(ConanException) as e:
            assert self.sut.optimized != 5
        assert "'5' is not a valid 'options.optimized' value" in str(e.value)

        with pytest.raises(ConanException) as e:
            self.sut.static = 1
        assert "'1' is not a valid 'options.static' value" in str(e.value)

        with pytest.raises(ConanException) as e:
            self.sut.optimized = 5
        assert "'5' is not a valid 'options.optimized' value" in str(e.value)

    def test_non_existing_option(self):
        with pytest.raises(ConanException) as e:
            assert self.sut.potato
        assert "option 'potato' doesn't exist" in str(e.value)

    def test_int(self):
        assert 3 == int(self.sut.optimized)

    def test_in(self):
        assert "static" in self.sut
        assert "potato" not in self.sut
        assert "optimized" in self.sut
        assert "path" in self.sut

    def test_assign(self):
        self.sut.static = False
        self.sut.optimized = 2
        self.sut.optimized = "4"
        self.sut.path = "otherpath"

        assert not self.sut.static
        assert self.sut.optimized == 4
        assert self.sut.path == "otherpath"

    def test_dumps(self):
        text = self.sut.dumps()
        # Output is ordereded alphabetically
        expected = textwrap.dedent("""\
            optimized=3
            path=mypath
            static=True""")
        assert text == expected

    def test_freeze(self):
        assert self.sut.static

        self.sut.freeze()
        # Should be freezed now
        # same value should not raise
        self.sut.static = True

        # Different value should raise
        with pytest.raises(ConanException) as e:
            self.sut.static = False
        assert "Incorrect attempt to modify option 'static'" in str(e.value)
        assert "static=True" in self.sut.dumps()

        # Removal of options with values doesn't raise anymore
        del self.sut.static
        assert "static" not in self.sut.dumps()

        # Test None is possible to change
        sut2 = Options({"static": [True, False],
                        "other": [True, False]})
        sut2.freeze()
        sut2.static = True
        assert "static=True" in sut2.dumps()
        # But not twice
        with pytest.raises(ConanException) as e:
            sut2.static = False
        assert "Incorrect attempt to modify option 'static'" in str(e.value)
        assert "static=True" in sut2.dumps()

        # can remove other, removing is always possible, even if freeze
        del sut2.other
        assert "other" not in sut2.dumps()

    def test_items(self):
        assert self.sut.items() == [("optimized", "3"), ("path", "mypath"), ("static", "True")]
        assert self.sut.items() == [("optimized", "3"), ("path", "mypath"), ("static", "True")]

    def test_get_safe_options(self):
        assert True == self.sut.get_safe("static")
        assert 3 == self.sut.get_safe("optimized")
        assert "mypath" == self.sut.get_safe("path")
        assert None == self.sut.get_safe("unknown")
        self.sut.path = "None"
        self.sut.static = False
        assert False == self.sut.get_safe("static")
        assert "None" == self.sut.get_safe("path")
        assert False == self.sut.get_safe("static", True)
        assert "None" == self.sut.get_safe("path", True)
        assert True == self.sut.get_safe("unknown", True)


class TestOptionsLoad:
    def test_load(self):
        text = textwrap.dedent("""\
            optimized=3
            path=mypath
            static=True
            zlib*:option=8
            *:common=value
            """)
        sut = Options.loads(text)
        assert sut.optimized == 3
        assert sut.optimized != "whatever"  # Non validating
        assert sut.path == "mypath"
        assert sut.path != "whatever"  # Non validating
        assert sut.static == "True"
        assert sut.static != "whatever"  # Non validating
        assert sut["zlib*"].option == 8
        assert sut["zlib*"].option != "whatever"  # Non validating
        assert sut["*"].common == "value"
        assert sut["*"].common != "whatever"  # Non validating


class TestOptionsPropagate:
    def test_basic(self):
        options = {"static": [True, False]}
        values = {"static": True}
        sut = Options(options, values)
        assert sut.static

        ref = RecipeReference.loads("boost/1.0")
        # if ref!=None option MUST be preceded by boost:
        down_options = Options(options_values={"zlib/2.0:other": 1, "boost/1.0:static": False})
        sut.apply_downstream(down_options, Options(), ref, False)
        assert not sut.static

        # Should be freezed now
        with pytest.raises(ConanException) as e:
            sut.static = True
        assert "Incorrect attempt to modify option 'static'" in str(e.value)

        self_options, up_options, up_private = sut.get_upstream_options(down_options, ref, False)
        assert up_options.dumps() == "zlib/2.0:other=1"
        assert self_options.dumps() == "boost/1.0:static=False\nzlib/2.0:other=1"
        assert up_private.dumps() == ""


class TestOptionsNone:
    @pytest.fixture(autouse=True)
    def _setup(self):
        options = {"static": [None, 1, 2], "other": [None, "ANY"], "more": ["None", 1]}
        self.sut = Options(options)

    def test_booleans(self):
        assert self.sut.static == None
        assert not self.sut.static
        assert self.sut.static != 1
        assert self.sut.static != 2
        with pytest.raises(ConanException) as e:
            self.sut.static == 3
        assert "'3' is not a valid 'options.static' value" in str(e.value)

        with pytest.raises(ConanException) as e:
            self.sut.static == "None"
        assert "'None' is not a valid 'options.static' value" in str(e.value)

        assert self.sut.other == None
        assert self.sut.other != "whatever"  # dont raise, ANY
        self.sut.other = None
        assert self.sut.other == None

        assert not self.sut.more
        assert self.sut.more == None
        assert self.sut.more != 1
        with pytest.raises(ConanException) as e:
            self.sut.more == 2
        assert "'2' is not a valid 'options.more' value" in str(e.value)
        with pytest.raises(ConanException) as e:
            self.sut.more = None
        assert "'None' is not a valid 'options.more' value" in str(e.value)
        self.sut.more = "None"
        assert not self.sut.more  # This is still evaluated to false, like OFF, 0, FALSE, etc
        assert self.sut.more == "None"
        assert self.sut.more != None

    def test_assign(self):
        self.sut.static = 1
        assert self.sut.static == 1

    def test_dumps(self):
        text = self.sut.dumps()
        assert text == ""

    def test_boolean_none(self):
        options = Options({"static": [None, "None", 1, 2]})
        assert options.static != 1
        assert not (options.static == 1)
        assert options.static != "None"
        assert not (options.static == "None")
        assert options.static == None
        assert not (options.static != None)

        options.static = "None"
        assert options.static == "None"
        assert not (options.static != "None")
        assert not (options.static == None)
        assert options.static != None

    def test_undefined_value(self):
        """ Not assigning a value to options will raise an error at validate() step
        """
        package_options = Options({"path": ["ANY"]})
        with pytest.raises(ConanException):
            package_options.validate()
        package_options.path = "Something"
        package_options.validate()

    def test_undefined_value_none(self):
        """ The value None is allowed as default, not necessary to default to it
        """
        package_options = Options({"path": [None, "Other"]})
        package_options.validate()
        package_options = Options({"path": ["None", "Other"]})
        with pytest.raises(ConanException):  # Literal "None" string not good to be undefined
            package_options.validate()



import os
import platform

import pytest

from conan.test.utils.tools import TestClient, GenConanfile


@pytest.mark.skipif(platform.system() != "Windows", reason="needs windows")
def test_fail_rmdir():
    client = TestClient()
    client.save({"conanfile.py": GenConanfile()})
    client.run("create . --name=mypkg --version=0.1 --user=lasote --channel=testing")
    build_folder = client.created_layout().build()
    f = open(os.path.join(build_folder, "myfile"), "wb")
    f.write(b"Hello world")
    client.run("install --requires=mypkg/0.1@lasote/testing --build=*", assert_error=True)
    assert "ERROR: Couldn't remove folder" in client.out



from bottle import request

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient, TestServer


class AuthorizationHeaderSpy:
    """ Generic plugin to handle Authorization header. Must be extended and implement
    some abstract methods in subclasses"""

    name = 'authorizationheaderspy'
    api = 2

    def __init__(self):
        self.auths = []

    def apply(self, callback, context):  # noqa
        auth = request.headers.get("Authorization")
        name = callback.__name__
        self.auths.append((name, auth))
        return callback


class TestAuthorizeBearer:

    def test_basic(self):
        auth = AuthorizationHeaderSpy()
        server = TestServer(plugins=[auth])
        client = TestClient(servers={"default": server}, inputs=["admin", "password"])
        client.save({"conanfile.py": GenConanfile("hello", "0.1")})
        client.run("export . --user=lasote --channel=stable")
        errors = client.run("upload hello/0.1@lasote/stable -r default --only-recipe")
        assert not errors

        expected_calls = [('get_recipe_revisions_references', None),
                          ('check_credentials', None),
                          ('authenticate', 'Basic'),
                          ('upload_recipe_file', 'Bearer')]

        assert len(expected_calls) == len(auth.auths)
        for i, (method, auth_type) in enumerate(expected_calls):
            real_call = auth.auths[i]
            assert method == real_call[0]
            if auth_type:
                assert auth_type in real_call[1]



import os
import tarfile

from conan.tools.files import unzip
from conan.test.utils.mocks import ConanFileMock
from conan.test.utils.test_files import temp_folder
from conan.internal.util.files import load, save


def test_unzip_can_xz():
    tmp_dir = temp_folder()
    file_path = os.path.join(tmp_dir, "a_file.txt")
    save(file_path, "my content!")
    txz = os.path.join(tmp_dir, "sample.tar.xz")
    with tarfile.open(txz, "w:xz") as tar:
        tar.add(file_path, "a_file.txt")

    dest_folder = temp_folder()
    unzip(ConanFileMock(), txz, dest_folder)
    content = load(os.path.join(dest_folder, "a_file.txt"))
    assert content == "my content!"


def test_unzip_can_gz():
    tmp_dir = temp_folder()
    file_path = os.path.join(tmp_dir, "a_file.txt")
    save(file_path, "my content!")
    gz = os.path.join(tmp_dir, "sample.tar.gz")
    with tarfile.open(gz, "w:gz") as tar:
        tar.add(file_path, "a_file.txt")

    dest_folder = temp_folder()
    unzip(ConanFileMock(), gz, dest_folder)
    content = load(os.path.join(dest_folder, "a_file.txt"))
    assert content == "my content!"



import json
import textwrap
import time
from collections import OrderedDict
from datetime import timedelta

from unittest.mock import patch

from conan.internal.api.remotes.localdb import LocalDB
from conan.test.utils.tools import TestClient, TestServer
from conan.test.utils.env import environment_update


class TestUser:

    def test_command_user_no_remotes(self):
        """ Test that proper error is reported when no remotes are defined and conan user is executed
        """
        client = TestClient()
        client.run("remote list-users", assert_error=True)
        assert "ERROR: No remotes defined" in client.out

        client.run("remote login wrong_remote foo -p bar", assert_error=True)
        assert "ERROR: Remote 'wrong_remote' can't be found or is disabled" in client.out

    def test_command_user_list(self):
        """ Test list of user is reported for all remotes or queried remote
        """
        servers = OrderedDict()
        servers["default"] = TestServer()
        servers["test_remote_1"] = TestServer()
        client = TestClient(servers=servers)

        # Test with wrong remote right error is reported
        client.run("remote login Test_Wrong_Remote foo", assert_error=True)
        assert "ERROR: Remote 'Test_Wrong_Remote' can't be found or is disabled" in client.out

        # Test user list for all remotes is reported
        client.run("remote list-users")
        assert textwrap.dedent("""
        default:
          No user
        test_remote_1:
          No user""") not in client.out

    def test_with_remote_no_connect(self):
        test_server = TestServer()
        client = TestClient(servers={"default": test_server})
        client.run('remote list-users')
        assert textwrap.dedent("""
                default:
                  No user""") not in client.out

        client.run('remote set-user default john')
        assert "Changed user of remote 'default' from 'None' (anonymous) to 'john'" in client.out
        localdb = LocalDB(client.cache_folder)
        assert ('john', None, None) == localdb.get_login(test_server.fake_url)

        client.run('remote set-user default will')
        assert "Changed user of remote 'default' from 'john' (anonymous) to 'will'" in client.out
        assert ('will', None, None) == localdb.get_login(test_server.fake_url)

        client.run('remote logout default')
        assert "Changed user of remote 'default' from 'will' (anonymous) to 'None' (anonymous)" in client.out
        assert (None, None, None) == localdb.get_login(test_server.fake_url)

    def test_command_user_with_password(self):
        """ Checks the -p option, that obtains a token from the password.
        Useful for integrations as travis, that interactive password is not
        possible
        """
        test_server = TestServer()
        servers = {"default": test_server}
        client = TestClient(servers=servers, inputs=["admin", "password"])
        client.run('remote login default dummy -p ping_pong2', assert_error=True)
        assert "ERROR: Wrong user or password" in client.out
        client.run('remote login default admin -p password')
        assert "ERROR: Wrong user or password" not in client.out
        assert "Changed user of remote 'default' from 'None' (anonymous) to 'admin'" in client.out
        client.run('remote logout default')
        assert "Changed user of remote 'default' from 'admin' (authenticated) to 'None' (anonymous)" in client.out
        localdb = LocalDB(client.cache_folder)
        assert (None, None, None) == localdb.get_login(test_server.fake_url)
        client.run('remote list-users')
        assert 'default:\n  No user' in client.out

    def test_command_user_with_password_spaces(self):
        """ Checks the -p option, that obtains a token from the password.
        Useful for integrations as travis, that interactive password is not
        possible
        """
        test_server = TestServer(users={"lasote": 'my "password'})
        servers = {"default": test_server}
        client = TestClient(servers=servers, inputs=["lasote", "mypass"])
        client.run(r'remote login default lasote -p="my \"password"')
        assert "Connecting to remote" not in client.out
        assert "Changed user of remote 'default' from 'None' (anonymous) to 'lasote'" in client.out
        client.run('remote logout default')
        client.run(r'remote login default lasote -p "my \"password"')
        assert "ERROR: Wrong user or password" not in client.out
        assert "Changed user of remote 'default' from 'None' (anonymous) to 'lasote'" in client.out

    def test_clean(self):
        test_server = TestServer()
        servers = {"default": test_server}
        client = TestClient(servers=servers, inputs=2*["admin", "password"])
        base = '''
from conan import ConanFile

class ConanLib(ConanFile):
    name = "lib"
    version = "0.1"
'''
        files = {"conanfile.py": base}
        client.save(files)
        client.run("export . --user=lasote --channel=stable")
        client.run("upload lib/0.1@lasote/stable -r default --only-recipe")
        client.run("remote list-users")
        assert 'default:\n  Username: admin\n  authenticated: True' in client.out
        client.run("remote logout default")
        assert "Changed user of remote 'default' from 'admin' (authenticated) to 'None' (anonymous)" in client.out
        client.run("remote list-users")
        assert 'default:\n  No user' in client.out
        # --force will force re-authentication, otherwise not necessary to auth
        client.run("upload lib/0.1@lasote/stable -r default --force --only-recipe")
        client.run("remote list-users")
        assert 'default:\n  Username: admin\n  authenticated: True' in client.out

    def test_command_interactive_only(self):
        test_server = TestServer()
        servers = {"default": test_server}
        client = TestClient(servers=servers, inputs=["password"])
        client.run('remote login default admin -p')
        assert "Changed user of remote 'default' from 'None' (anonymous) to 'admin' (authenticated)" in client.out

    def test_command_user_with_interactive_password_login_prompt_disabled(self):
        """ Interactive password should not work.
        """
        test_server = TestServer()
        servers = {"default": test_server}
        client = TestClient(servers=servers,  inputs=[])
        conan_conf = "core:non_interactive=True"
        client.save_home({"global.conf": conan_conf})
        client.run('remote login default admin -p', assert_error=True)
        assert 'ERROR: Conan interactive mode disabled' in client.out
        assert "Please enter a password for user 'admin'" not in client.out
        client.run("remote list-users")
        assert "default:\n  No user" in client.out

    def test_authenticated(self):
        test_server = TestServer(users={"lasote": "mypass", "danimtb": "passpass"})
        servers = OrderedDict()
        servers["default"] = test_server
        servers["other_server"] = TestServer()
        client = TestClient(servers=servers, inputs=["lasote", "mypass", "mypass", "mypass"])
        client.run("remote logout default")
        assert "Changed user of remote 'default' from 'None' (anonymous) to 'None' (anonymous)" in client.out
        assert "[authenticated]" not in client.out
        client.run('remote set-user default bad_user')
        client.run("remote list-users")
        assert 'default:\n  Username: bad_user\n  authenticated: False' in client.out
        client.run("remote set-user default lasote")
        client.run("remote list-users")
        assert 'default:\n  Username: lasote\n  authenticated: False' in client.out
        client.run("remote login default lasote -p mypass")
        client.run("remote list-users")
        assert 'default:\n  Username: lasote\n  authenticated: True' in client.out

        client.run("remote login default danimtb -p passpass")
        assert "Changed user of remote 'default' from 'lasote' (authenticated) to 'danimtb' (authenticated)" in client.out
        client.run("remote list-users")
        assert 'default:\n  Username: danimtb\n  authenticated: True' in client.out

    def test_json(self):
        default_server = TestServer(users={"lasote": "mypass", "danimtb": "passpass"})
        other_server = TestServer()
        servers = OrderedDict()
        servers["default"] = default_server
        servers["other_server"] = other_server
        client = TestClient(servers=servers, inputs=["lasote", "mypass", "danimtb", "passpass"])
        client.run("remote list-users -f json")
        info = json.loads(client.stdout)
        assert info == [
                            {
                                "name": "default",
                                "authenticated": False,
                                "user_name": None
                            },
                            {
                                "name": "other_server",
                                "authenticated": False,
                                "user_name": None
                            }
                        ]

        client.run('remote set-user default bad_user')
        client.run("remote list-users -f json")
        info = json.loads(client.stdout)
        assert info == [
            {
                "name": "default",
                "authenticated": False,
                "user_name": "bad_user"
            },
            {
                "name": "other_server",
                "authenticated": False,
                "user_name": None
            }
        ]

        client.run('remote set-user default lasote')
        client.run("remote list-users -f json")
        info = json.loads(client.stdout)
        assert info == [
            {
                "name": "default",
                "authenticated": False,
                "user_name": "lasote"
            },
            {
                "name": "other_server",
                "authenticated": False,
                "user_name": None
            }
        ]

        client.run("remote login default lasote -p mypass")
        client.run("remote list-users -f json")
        info = json.loads(client.stdout)
        assert info == [
            {
                "name": "default",
                "authenticated": True,
                "user_name": "lasote"
            },
            {
                "name": "other_server",
                "authenticated": False,
                "user_name": None
            }
        ]

        client.run("remote login default danimtb -p passpass")
        client.run("remote list-users -f json")
        info = json.loads(client.stdout)
        assert info == [
            {
                "name": "default",
                "authenticated": True,
                "user_name": "danimtb"
            },
            {
                "name": "other_server",
                "authenticated": False,
                "user_name": None
            }
        ]
        client.run("remote set-user other_server lasote")
        client.run("remote list-users -f json")
        info = json.loads(client.stdout)
        assert info == [
            {
                "name": "default",
                "authenticated": True,
                "user_name": "danimtb"
            },
            {
                "name": "other_server",
                "authenticated": False,
                "user_name": "lasote"
            }
        ]

        client.run("remote logout '*'")
        client.run("remote set-user default danimtb")
        client.run("remote list-users -f json")
        info = json.loads(client.stdout)
        assert info == [
            {
                "name": "default",
                "authenticated": False,
                "user_name": "danimtb"
            },
            {
                "name": "other_server",
                "authenticated": False,
                "user_name": None
            }
        ]
        client.run("remote list-users")
        assert "default:\n  Username: danimtb\n  authenticated: False" in client.out
        assert "other_server:\n  No user\n" in client.out

    def test_skip_auth(self):
        default_server = TestServer(users={"lasote": "mypass", "danimtb": "passpass"})
        servers = OrderedDict()
        servers["default"] = default_server
        client = TestClient(servers=servers)
        # Regular auth
        client.run("remote login default lasote -p mypass")

        # Now skip the auth but keeping the same user
        client.run("remote set-user default lasote")
        assert "Changed user of remote 'default' from 'lasote' (authenticated) to 'lasote' (authenticated)" in client.out

        # If we change the user the credentials are removed
        client.run("remote set-user default flanders")
        assert "Changed user of remote 'default' from 'lasote' (authenticated) to 'flanders' (anonymous)" in client.out

        client.run("remote login default lasote -p BAD_PASS", assert_error=True)
        assert "Wrong user or password" in client.out

        # Login again correctly
        client.run("remote login default lasote -p mypass")

    def test_login_multiremote(self):
        servers = OrderedDict()
        servers["default"] = TestServer(users={"admin": "password"})
        servers["other"] = TestServer(users={"admin": "password"})
        c = TestClient(servers=servers, inputs=["admin", "password", "wrong", "wrong"])
        # This must fail, not autthenticate in the next remote
        c.run("remote login *", assert_error=True)
        assert "ERROR: Wrong user or password. [Remote: other]" in c.out


def test_user_removed_remote_removed():
    # Make sure that removing a remote clears the credentials
    # https://github.com/conan-io/conan/issues/5562
    c = TestClient(default_server_user=True)
    server_url = c.servers["default"].fake_url
    c.run("remote login default admin -p password")
    localdb = LocalDB(c.cache_folder)
    login = localdb.get_login(server_url)
    assert login[0] == "admin"
    c.run("remote remove default")
    login = localdb.get_login(server_url)
    assert login == (None, None, None)


class TestRemoteAuth:
    def test_remote_auth(self):
        servers = OrderedDict()
        servers["default"] = TestServer(users={"lasote": "mypass", "danimtb": "passpass"})
        servers["other_server"] = TestServer(users={"lasote": "mypass"})
        c = TestClient(servers=servers, inputs=["lasote", "mypass", "danimtb", "passpass",
                                                "lasote", "mypass"])
        c.run("remote auth *")
        text = textwrap.dedent("""\
            default:
                user: lasote
            other_server:
                user: lasote""")
        assert text in c.out

        c.run("remote auth * --format=json")
        result = json.loads(c.stdout)
        assert result == {'default': {'user': 'lasote'}, 'other_server': {'user': 'lasote'}}

    def test_remote_auth_force(self):
        servers = OrderedDict()
        servers["default"] = TestServer(users={"lasote": "mypass", "danimtb": "passpass"})
        servers["other_server"] = TestServer(users={"lasote": "mypass"})
        c = TestClient(servers=servers, inputs=["lasote", "mypass", "danimtb", "passpass",
                                                "lasote", "mypass"])

        with patch("conan.internal.rest.rest_client_v2.RestV2Methods.check_credentials") as check_credentials_mock:
            c.run("remote auth --force *")
            check_credentials_mock.assert_called_with(True)

    def test_remote_auth_force_false(self):
        servers = OrderedDict()
        servers["default"] = TestServer(users={"lasote": "mypass", "danimtb": "passpass"})
        servers["other_server"] = TestServer(users={"lasote": "mypass"})
        c = TestClient(servers=servers, inputs=["lasote", "mypass", "danimtb", "passpass",
                                                "lasote", "mypass"])

        with patch("conan.internal.rest.rest_client_v2.RestV2Methods.check_credentials") as check_credentials_mock:
            c.run("remote auth *")
            check_credentials_mock.assert_called_with(False)

    def test_remote_auth_with_user(self):
        servers = OrderedDict()
        servers["default"] = TestServer(users={"lasote": "mypass"})
        servers["other_server"] = TestServer()
        c = TestClient(servers=servers, inputs=["lasote", "mypass"])
        c.run("remote set-user default lasote")
        c.run("remote auth * --with-user")
        text = textwrap.dedent("""\
            default:
                user: lasote
            other_server:
                user: None""")
        assert text in c.out

    def test_remote_auth_with_user_env_var(self):
        servers = OrderedDict()
        servers["default"] = TestServer(users={"lasote": "mypass"})
        servers["other_server"] = TestServer()
        c = TestClient(servers=servers)
        with environment_update({"CONAN_LOGIN_USERNAME_DEFAULT": "lasote",
                                 "CONAN_PASSWORD_DEFAULT": "mypass"}):
            c.run("remote auth * --with-user")
        text = textwrap.dedent("""\
            default:
                user: lasote
            other_server:
                user: None""")
        assert text in c.out

    def test_remote_auth_error(self):
        servers = OrderedDict()
        servers["default"] = TestServer(users={"user": "password"})
        c = TestClient(servers=servers, inputs=["user1", "pass", "user2", "pass", "user3", "pass"])
        c.run("remote auth *")
        assert "error: Too many failed login attempts, bye!" in c.out

    def test_remote_auth_server_expire_token_secret(self):
        server = TestServer(users={"myuser": "password", "myotheruser": "otherpass"})
        c = TestClient(servers={"default": server}, inputs=["myuser", "password",
                                                            "myotheruser", "otherpass",
                                                            "user", "pass", "user", "pass",
                                                            "user", "pass"])
        c.run("remote auth *")
        assert "Remote 'default' needs authentication, obtaining credentials" in c.out
        assert "user: myuser" in c.out
        c.run("remote auth *")
        assert "Remote 'default' needs authentication, obtaining credentials" not in c.out
        assert "user: myuser" in c.out
        # Invalidate server secret
        server.test_server.ra.api_v2.credentials_manager.secret = "potato"
        c.run("remote auth *")
        assert "user: myotheruser" in c.out
        # Invalidate server secret again
        server.test_server.ra.api_v2.credentials_manager.secret = "potato2"
        c.run("remote auth *")
        assert "error: Too many failed login attempts, bye!" in c.out

    def test_remote_auth_server_expire_token(self):
        server = TestServer(users={"myuser": "password", "myotheruser": "otherpass"})
        server.test_server.ra.api_v2.credentials_manager.expire_time = timedelta(seconds=2)
        c = TestClient(servers={"default": server}, inputs=["myuser", "password",
                                                            "myotheruser", "otherpass",
                                                            "user", "pass", "user", "pass",
                                                            "user", "pass"])
        c.run("remote auth *")
        assert "user: myuser" in c.out
        # token not expired yet, should work
        c.run("remote auth *")
        assert "user: myuser" in c.out
        # Token should expire
        time.sleep(3)
        c.run("remote auth *")
        assert "user: myotheruser" in c.out
        # Token should expire
        time.sleep(3)
        c.run("remote auth *")
        assert "error: Too many failed login attempts, bye!" in c.out

    def test_auth_after_logout(self):
        server = TestServer(users={"myuser": "password"})
        c = TestClient(servers={"default": server}, inputs=["myuser", "password"]*2)
        c.run("remote auth *")
        assert "Remote 'default' needs authentication, obtaining credentials" in c.out
        assert "user: myuser" in c.out
        c.run("remote logout *")
        assert "Changed user of remote 'default' from 'myuser' (authenticated) to 'None'" in c.out
        c.run("remote auth *")
        assert "Remote 'default' needs authentication, obtaining credentials" in c.out
        assert "user: myuser" in c.out



import json
import os
import shutil
import stat
import textwrap

import pytest
from unittest.mock import patch

import yaml

from conan.api.model import Remote, RecipeReference
from conan.internal.api.config.config_installer import _hide_password

from conan.internal.cache.home_paths import HomePaths
from conan.internal.rest.file_downloader import FileDownloader
from conan.internal.paths import DEFAULT_CONAN_HOME
from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.file_server import TestFileServer
from conan.test.utils.test_files import scan_folder, temp_folder, tgz_with_contents
from conan.test.utils.tools import TestClient, zipdir
from conan.internal.util.files import load, mkdir, save, save_files


def make_file_read_only(file_path):
    mode = os.stat(file_path).st_mode
    os.chmod(file_path, mode & ~ stat.S_IWRITE)


remotes = """{
 "remotes": [
  {
   "name": "myrepo1",
   "url": "https://myrepourl.net",
   "verify_ssl": false
  },
  {
   "name": "my-repo-2",
   "url": "https://myrepo2.com",
   "verify_ssl": true
  }
 ]
}
"""

settings_yml = """os:
    Windows:
    Linux:
arch: [x86, x86_64]
"""


class TestConfigInstall:

    @staticmethod
    def _create_profile_folder(folder=None):
        folder = folder or temp_folder(path_with_spaces=False)
        save_files(folder, {"settings.yml": settings_yml,
                            "remotes.json": remotes,
                            "profiles/linux": "#linuxprofile",
                            "profiles/windows": "#winprofile",
                            "hooks/dummy": "#hook dummy",
                            "hooks/foo.py": "#hook foo",
                            "hooks/custom/custom.py": "#hook custom",
                            ".git/hooks/foo": "foo",
                            "hooks/.git/hooks/before_push": "before_push",
                            "pylintrc": "#Custom pylint",
                            "python/myfuncs.py": "does not matter",
                            "python/__init__.py": ""
                            })
        return folder

    def test_config_fails_no_storage(self):
        folder = temp_folder(path_with_spaces=False)
        save_files(folder, {"remotes.json": remotes})
        client = TestClient()
        client.save({"conanfile.py": GenConanfile()})
        client.run("create . --name=pkg --version=1.0")
        client.run('config install "%s"' % folder)
        client.run("remote list")
        assert "myrepo1: https://myrepourl.net [Verify SSL: False, Enabled: True]" in client.out
        assert "my-repo-2: https://myrepo2.com [Verify SSL: True, Enabled: True]" in client.out

    def _create_zip(self, zippath=None):
        folder = self._create_profile_folder()
        zippath = zippath or os.path.join(folder, "myconfig.zip")
        zipdir(folder, zippath)
        return zippath

    @staticmethod
    def _get_files(folder):
        relpaths = scan_folder(folder)
        files = {}
        for path in relpaths:
            with open(os.path.join(folder, path), "r") as file_handle:
                files[path] = file_handle.read()
        return files

    def _create_tgz(self, tgz_path=None):
        folder = self._create_profile_folder()
        tgz_path = tgz_path or os.path.join(folder, "myconfig.tar.gz")
        files = self._get_files(folder)
        return tgz_with_contents(files, tgz_path)

    @staticmethod
    def _check(c):
        settings_path = c.paths.settings_path
        assert load(settings_path).splitlines() == settings_yml.splitlines()
        api = c.api
        cache_remotes = api.remotes.list()
        assert list(cache_remotes) == [
            Remote("myrepo1", "https://myrepourl.net", False, False),
            Remote("my-repo-2", "https://myrepo2.com", True, False),
        ]
        assert sorted(os.listdir(c.paths.profiles_path)) == sorted(["default", "linux", "windows"])
        assert c.load_home("profiles/linux") == "#linuxprofile"
        assert c.load_home("profiles/windows") == "#winprofile"
        assert "#Custom pylint" == c.load_home("pylintrc")
        assert "" == c.load_home("python/__init__.py")
        assert "#hook dummy" == c.load_home("hooks/dummy")
        assert "#hook foo" == c.load_home("hooks/foo.py")
        assert "#hook custom" == c.load_home("hooks/custom/custom.py")
        assert not os.path.exists(os.path.join(c.cache_folder, "hooks", ".git"))
        assert not os.path.exists(os.path.join(c.cache_folder, ".git"))

    def test_install_file(self):
        """ should install from a file in current dir
        """
        zippath = self._create_zip()
        c = TestClient(light=True)
        for filetype in ["", "--type=file"]:
            c.run('config install "%s" %s' % (zippath, filetype))
            self._check(c)
            assert os.path.exists(zippath)

    def test_install_config_file(self):
        """ should install from a settings and remotes file in configuration directory
        """
        import tempfile
        profile_folder = self._create_profile_folder()
        assert os.path.isdir(profile_folder)
        src_setting_file = os.path.join(profile_folder, "settings.yml")
        src_remote_file = os.path.join(profile_folder, "remotes.json")

        # Install profile_folder without settings.yml remotes.json in order to install them manually
        tmp_dir = tempfile.mkdtemp()
        dest_setting_file = os.path.join(tmp_dir, "settings.yml")
        dest_remote_file = os.path.join(tmp_dir, "remotes.json")
        shutil.move(src_setting_file, dest_setting_file)
        shutil.move(src_remote_file, dest_remote_file)

        c = TestClient(light=True)
        c.run('config install "%s"' % profile_folder)
        shutil.move(dest_setting_file, src_setting_file)
        shutil.move(dest_remote_file, src_remote_file)
        shutil.rmtree(tmp_dir)

        for cmd_option in ["", "--type=file"]:
            c.run('config install "%s" %s' % (src_setting_file, cmd_option))
            c.run('config install "%s" %s' % (src_remote_file, cmd_option))
            self._check(c)

    def test_install_dir(self):
        """ should install from a dir in current dir
        """
        folder = self._create_profile_folder()
        assert os.path.isdir(folder)
        c = TestClient(light=True)
        for dirtype in ["", "--type=dir"]:
            c.run('config install "%s" %s' % (folder, dirtype))
            self._check(c)

    def test_install_source_target_folders(self):
        folder = temp_folder()
        save_files(folder, {"subf/file.txt": "hello",
                            "subf/subf/file2.txt": "bye"})
        c = TestClient(light=True)
        c.run('config install "%s" -sf=subf -tf=newsubf' % folder)
        content = c.load_home("newsubf/file.txt")
        assert content == "hello"
        content = c.load_home("newsubf/subf/file2.txt")
        assert content == "bye"

    def test_install_remotes_json(self):
        folder = temp_folder()

        remotes_json = textwrap.dedent("""
            {
                "remotes": [
                    { "name": "repojson1", "url": "https://repojson1.net", "verify_ssl": false },
                    { "name": "repojson2", "url": "https://repojson2.com", "verify_ssl": true }
                ]
            }
        """)

        remotes_txt = textwrap.dedent("""\
            repotxt1 https://repotxt1.net False
            repotxt2 https://repotxt2.com True
        """)

        # remotes.txt is ignored
        save_files(folder, {"remotes.json": remotes_json,
                            "remotes.txt": remotes_txt})

        c = TestClient(light=True)
        c.run(f'config install "{folder}"')
        assert "Defining remotes from remotes.json" in c.out

        c.run('remote list')

        assert "repojson1: https://repojson1.net [Verify SSL: False, Enabled: True]" in c.out
        assert "repojson2: https://repojson2.com [Verify SSL: True, Enabled: True]" in c.out

        # We only install remotes.json
        folder = temp_folder()
        save_files(folder, {"remotes.json": remotes_json})

        c.run(f'config install "{folder}"')
        assert "Defining remotes from remotes.json" in c.out

        c.run('remote list')

        assert "repojson1: https://repojson1.net [Verify SSL: False, Enabled: True]" in c.out
        assert "repojson2: https://repojson2.com [Verify SSL: True, Enabled: True]" in c.out

    def test_without_profile_folder(self):
        c = TestClient(light=True)
        shutil.rmtree(c.paths.profiles_path)
        zippath = self._create_zip()

        c.run('config install "%s"' % zippath)
        assert sorted(os.listdir(c.paths.profiles_path)) == sorted(["linux", "windows"])
        assert c.load_home("profiles/linux") == "#linuxprofile"

    def test_install_url(self):
        """ should install from a URL
        """
        c = TestClient(light=True)
        for origin in ["", "--type=url"]:
            def my_download(obj, url, file_path, **kwargs):  # noqa
                self._create_zip(file_path)

            with patch.object(FileDownloader, 'download', new=my_download):
                c.run("config install http://myfakeurl.com/myconf.zip %s" % origin)
                self._check(c)

                # repeat the process to check
                c.run("config install http://myfakeurl.com/myconf.zip %s" % origin)
                self._check(c)

    def test_install_url_query(self):
        """ should install from a URL
        """
        c = TestClient(light=True)

        def my_download(obj, url, file_path, **kwargs):  # noqa
            self._create_zip(file_path)

        with patch.object(FileDownloader, 'download', new=my_download):
            # repeat the process to check it works with ?args
            c.run("config install http://myfakeurl.com/myconf.zip?sha=1")
            self._check(c)

    def test_install_change_only_verify_ssl(self):
        def my_download(obj, url, file_path, **kwargs):  # noqa
            self._create_zip(file_path)

        c = TestClient(light=True)
        with patch.object(FileDownloader, 'download', new=my_download):
            c.run("config install http://myfakeurl.com/myconf.zip")
            self._check(c)

            # repeat the process to check
            c.run("config install http://myfakeurl.com/myconf.zip --verify-ssl=False")
            self._check(c)

    def test_install_url_tgz(self):
        """ should install from a URL to tar.gz
        """
        c = TestClient(light=True)

        def my_download(obj, url, file_path, **kwargs):  # noqa
            self._create_tgz(file_path)

        with patch.object(FileDownloader, 'download', new=my_download):
            c.run("config install http://myfakeurl.com/myconf.tar.gz")
            self._check(c)

    def test_failed_install_repo(self):
        """ should install from a git repo
        """
        c = TestClient(light=True)
        c.run('config install notexistingrepo.git', assert_error=True)
        assert "ERROR: Failed conan config install: Can't clone repo" in c.out

    def test_failed_install_http(self):
        """ should install from a http zip
        """
        c = TestClient(light=True)
        c.run('config install httpnonexisting', assert_error=True)
        assert ("ERROR: Failed conan config install: "
                "Error while installing config from httpnonexisting") in c.out

    @pytest.mark.tool("git")
    def test_install_repo(self):
        """ should install from a git repo
        """
        c = TestClient(light=True)
        folder = self._create_profile_folder()
        with c.chdir(folder):
            c.run_command('git init .')
            c.run_command('git add .')
            c.run_command('git config user.name myname')
            c.run_command('git config user.email myname@mycompany.com')
            c.run_command('git commit -m "mymsg"')

        c.run('config install "%s/.git"' % folder)
        self._check(c)

    @pytest.mark.tool("git")
    def test_install_repo_relative(self):
        c = TestClient(light=True)
        relative_folder = "./config"
        absolute_folder = os.path.join(c.current_folder, "config")
        mkdir(absolute_folder)
        folder = self._create_profile_folder(absolute_folder)
        with c.chdir(folder):
            c.run_command('git init .')
            c.run_command('git add .')
            c.run_command('git config user.name myname')
            c.run_command('git config user.email myname@mycompany.com')
            c.run_command('git commit -m "mymsg"')

        c.run('config install "%s/.git"' % relative_folder)
        self._check(c)

    @pytest.mark.tool("git")
    def test_install_custom_args(self):
        """ should install from a git repo
        """
        c = TestClient(light=True)
        folder = self._create_profile_folder()
        with c.chdir(folder):
            c.run_command('git init .')
            c.run_command('git add .')
            c.run_command('git config user.name myname')
            c.run_command('git config user.email myname@mycompany.com')
            c.run_command('git commit -m "mymsg"')

        c.run('config install "%s/.git" --args="-c init.templateDir=value"' % folder)
        self._check(c)

    def test_force_git_type(self):
        client = TestClient(light=True)
        client.run('config install httpnonexisting --type=git', assert_error=True)
        assert "Can't clone repo" in client.out

    def test_force_dir_type(self):
        c = TestClient(light=True)
        c.run('config install httpnonexisting --type=dir', assert_error=True)
        assert "ERROR: Failed conan config install: No such directory: 'httpnonexisting'" in c.out

    def test_force_file_type(self):
        client = TestClient(light=True)
        client.run('config install httpnonexisting --type=file', assert_error=True)
        assert "No such file or directory: 'httpnonexisting'" in client.out

    def test_force_url_type(self):
        client = TestClient(light=True)
        client.run('config install httpnonexisting --type=url', assert_error=True)
        assert "Error downloading file httpnonexisting: 'Invalid URL 'httpnonexisting'" in client.out

    def test_removed_credentials_from_url_unit(self):
        """
        Unit tests to remove credentials in netloc from url when using basic auth
        # https://github.com/conan-io/conan/issues/2324
        """
        url_without_credentials = r"https://server.com/resource.zip"
        url_with_credentials = r"https://test_username:test_password_123@server.com/resource.zip"
        url_hidden_password = r"https://test_username:<hidden>@server.com/resource.zip"

        # Check url is the same when not using credentials
        assert _hide_password(url_without_credentials) == url_without_credentials

        # Check password is hidden using url with credentials
        assert _hide_password(url_with_credentials) == url_hidden_password

        # Check that it works with other protocols ftp
        ftp_with_credentials = r"ftp://test_username_ftp:test_password_321@server.com/resurce.zip"
        ftp_hidden_password = r"ftp://test_username_ftp:<hidden>@server.com/resurce.zip"
        assert _hide_password(ftp_with_credentials) == ftp_hidden_password

        # Check function also works for file paths *unix/windows
        unix_file_path = r"/tmp/test"
        assert _hide_password(unix_file_path) == unix_file_path
        windows_file_path = r"c:\windows\test"
        assert _hide_password(windows_file_path) == windows_file_path

        # Check works with empty string
        assert _hide_password('') == ''

    def test_remove_credentials_config_installer(self):
        """ Functional test to check credentials are not displayed in output but are still present
        in conan configuration
        # https://github.com/conan-io/conan/issues/2324
        """
        fake_url_with_credentials = "http://test_user:test_password@myfakeurl.com/myconf.zip"
        fake_url_hidden_password = "http://test_user:<hidden>@myfakeurl.com/myconf.zip"

        def my_download(obj, url, file_path, **kwargs):  # noqa
            assert url == fake_url_with_credentials
            self._create_zip(file_path)

        c = TestClient(light=True)
        with patch.object(FileDownloader, 'download', new=my_download):
            c.run("config install %s" % fake_url_with_credentials)

            # Check credentials are not displayed in output
            assert fake_url_with_credentials not in c.out
            assert fake_url_hidden_password in c.out

            # Check credentials still stored in configuration
            self._check(c)

    def test_ssl_verify(self):
        c = TestClient(light=True)

        fake_url = "https://fakeurl.com/myconf.zip"

        def download_verify_false(obj, url, file_path, **kwargs):  # noqa
            assert kwargs["verify_ssl"] is False
            self._create_zip(file_path)

        def download_verify_true(obj, url, file_path, **kwargs):  # noqa
            assert kwargs["verify_ssl"] is True
            self._create_zip(file_path)

        with patch.object(FileDownloader, 'download', new=download_verify_false):
            c.run("config install %s --verify-ssl=False" % fake_url)

        with patch.object(FileDownloader, 'download', new=download_verify_true):
            c.run("config install %s --verify-ssl=True" % fake_url)

        with patch.object(FileDownloader, 'download', new=download_verify_true):
            c.run(f"config install {fake_url}")

        with patch.object(FileDownloader, 'download', new=download_verify_false):
            c.run(f"config install {fake_url} --insecure")

    @pytest.mark.tool("git")
    def test_git_checkout_is_possible(self):
        folder = self._create_profile_folder()
        c = TestClient(light=True)
        with c.chdir(folder):
            c.run_command('git init .')
            c.run_command('git checkout -b master')
            c.run_command('git add .')
            c.run_command('git config user.name myname')
            c.run_command('git config user.email myname@mycompany.com')
            c.run_command('git commit -m "mymsg"')
            c.run_command('git checkout -b other_branch')
            save(os.path.join(folder, "extensions", "hooks", "cust", "cust.py"), "")
            c.run_command('git add .')
            c.run_command('git commit -m "my file"')

        c.run('config install "%s/.git" --args "-b other_branch"' % folder)
        self._check(c)
        file_path = os.path.join(c.paths.hooks_path, "cust", "cust.py")
        assert load(file_path) == ""

        # Add changes to that branch and update
        with c.chdir(folder):
            save(os.path.join(folder, "extensions", "hooks", "cust", "cust.py"), "new content")
            c.run_command('git add .')
            c.run_command('git commit -m "my other file"')
            c.run_command('git checkout master')
        c.run('config install "%s/.git" --args "-b other_branch"' % folder)
        self._check(c)
        assert load(file_path) == "new content"

    def test_config_install_requester(self):
        # https://github.com/conan-io/conan/issues/4169
        path = self._create_zip()
        c = TestClient(light=True)
        file_server = TestFileServer(os.path.dirname(path))
        c.servers["file_server"] = file_server

        c.run(f"config install {file_server.fake_url}/myconfig.zip")
        assert "Defining remotes from remotes.json" in c.out
        assert "Copying file myfuncs.py" in c.out

    def test_overwrite_read_only_file(self):
        c = TestClient(light=True)
        source_folder = self._create_profile_folder()
        c.run('config install "%s"' % source_folder)
        # make existing settings.yml read-only
        make_file_read_only(c.paths.settings_path)
        assert not os.access(c.paths.settings_path, os.W_OK)

        # config install should overwrite the existing read-only file
        c.run('config install "%s"' % source_folder)
        assert os.access(c.paths.settings_path, os.W_OK)

    def test_dont_copy_file_permissions(self):
        source_folder = self._create_profile_folder()
        # make source settings.yml read-only
        make_file_read_only(os.path.join(source_folder, 'remotes.json'))

        c = TestClient(light=True)
        c.run('config install "%s"' % source_folder)
        assert os.access(c.paths.settings_path, os.W_OK)


class TestConfigInstallSched:

    def test_execute_more_than_once(self):
        """ Once executed by the scheduler, conan config install must executed again
            when invoked manually
        """
        folder = temp_folder(path_with_spaces=False)
        save_files(folder, {"global.conf": "core.download:parallel=0"})
        c = TestClient(light=True)
        c.run('config install "%s"' % folder)
        assert "Copying file global.conf" in c.out

        c.run('config install "%s"' % folder)
        assert "Copying file global.conf" in c.out

    @pytest.mark.tool("git")
    def test_config_install_remove_git_repo(self):
        """ config_install_interval must break when remote git has been removed
        """
        folder = temp_folder(path_with_spaces=False)
        save_files(folder, {"global.conf": "core.download:parallel=0"})
        c = TestClient(light=True)
        with c.chdir(folder):
            c.run_command('git init .')
            c.run_command('git add .')
            c.run_command('git config user.name myname')
            c.run_command('git config user.email myname@mycompany.com')
            c.run_command('git commit -m "mymsg"')
        c.run('config install "%s/.git" --type git' % folder)
        assert "Copying file global.conf" in c.out
        assert "Repo cloned!" in c.out  # git clone executed by scheduled task

    def test_config_fails_git_folder(self):
        # https://github.com/conan-io/conan/issues/8594
        config_folder = temp_folder(path_with_spaces=False)
        save_files(config_folder, {"global.conf": "core.download:parallel=0"})
        folder = os.path.join(temp_folder(), ".gitlab-conan", DEFAULT_CONAN_HOME)
        client = TestClient(cache_folder=folder)
        with client.chdir(config_folder):
            client.run_command('git init .')
            client.run_command('git add .')
            client.run_command('git config user.name myname')
            client.run_command('git config user.email myname@mycompany.com')
            client.run_command('git commit -m "mymsg"')
        assert ".gitlab-conan" in client.cache_folder
        assert os.path.basename(client.cache_folder) == DEFAULT_CONAN_HOME
        client.run('config install "%s/.git" --type git' % config_folder)
        client.load_home("global.conf")  # check it is there
        dirs = os.listdir(client.cache_folder)
        assert ".git" not in dirs


class TestConfigInstall2:
    def test_config_install_reestructuring_source(self):
        """  https://github.com/conan-io/conan/issues/9885 """

        folder = temp_folder()
        client = TestClient()
        with client.chdir(folder):
            client.save({"profiles/debug/address-sanitizer": ""})
            client.run("config install .")

        debug_cache_folder = os.path.join(client.cache_folder, "profiles", "debug")
        assert os.path.isdir(debug_cache_folder)

        # Now reestructure the files, what it was already a directory in the cache now we want
        # it to be a file
        folder = temp_folder()
        with client.chdir(folder):
            client.save({"profiles/debug": ""})
            client.run("config install .")
        assert os.path.isfile(debug_cache_folder)

        # And now is a directory again
        folder = temp_folder()
        with client.chdir(folder):
            client.save({"profiles/debug/address-sanitizer": ""})
            client.run("config install .")
        assert os.path.isdir(debug_cache_folder)


class TestConfigInstallPkg:

    @pytest.fixture(scope="class")
    def servers(self):
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            from conan.tools.files import copy
            class Conf(ConanFile):
                package_type = "configuration"
                def package(self):
                    copy(self, "*.conf", src=self.build_folder, dst=self.package_folder)
            """)
        c = TestClient(default_server_user=True, light=True)
        c.save({"conanfile.py": conanfile})
        for pkg in ("myconf_a", "myconf_b", "myconf_c"):
            for version in ("0.1", "0.2"):
                c.save({"global.conf": f"user.myteam:myconf=my_{pkg}/{version}_value"})
                c.run(f"export-pkg . --name={pkg} --version={version}")
        c.run("upload * -r=default -c")
        c.run("remove * -c")
        return c.servers

    @staticmethod
    def _check_conf(c, pkg):
        c.run("config show *")
        assert f"user.myteam:myconf: my_{pkg}_value" in c.out

    @staticmethod
    def _check_conf_file(c, refs):
        content = json.loads(c.load_home("config_version.json"))["config_version"]
        for a, b in zip(refs, content):
            assert a in b

    def test_install_pkg(self, servers):
        c = TestClient(servers=servers, light=True)
        c.run("config install-pkg myconf_a/0.1")
        assert "Installing new or updating configuration packages" in c.out
        self._check_conf(c, "myconf_a/0.1")
        self._check_conf_file(c, ["myconf_a/0.1"])
        cache_files = os.listdir(c.cache_folder)
        assert "conaninfo.txt" not in cache_files
        assert "conanmanifest.txt" not in cache_files

        # skip reinstall
        c.run("config install-pkg myconf_a/0.1")
        assert ("The requested configurations are identical to the already installed ones, "
                "skipping re-installation") in c.out
        self._check_conf(c, "myconf_a/0.1")
        self._check_conf_file(c, ["myconf_a/0.1"])

        # forced reinstall
        c.save_home({"global.conf": ""})
        c.run("config install-pkg myconf_a/0.1 --force")
        assert "forcing re-installation because --force" in c.out
        self._check_conf(c, "myconf_a/0.1")
        self._check_conf_file(c, ["myconf_a/0.1"])

    def test_with_url(self, servers):
        c = TestClient(servers=servers, light=True)
        url = servers["default"].fake_url
        c.run("remote remove default")
        c.run(f"config install-pkg myconf_a/0.1 --url={url}")
        assert "Installing new or updating configuration packages" in c.out
        self._check_conf(c, "myconf_a/0.1")
        self._check_conf_file(c, ["myconf_a/0.1"])

    def test_update(self, servers):
        c = TestClient(servers=servers, light=True)
        c.run("config install-pkg myconf_a/0.1")
        c.run("config install-pkg myconf_a/0.2")
        assert "Installing new or updating configuration packages" in c.out
        self._check_conf(c, "myconf_a/0.2")
        self._check_conf_file(c, ["myconf_a/0.2"])

    def test_addition(self, servers):
        c = TestClient(servers=servers, light=True)
        c.run("config install-pkg myconf_a/0.1")
        c.run("config install-pkg myconf_b/0.1")
        assert "Installing new or updating configuration packages" in c.out
        self._check_conf(c, "myconf_b/0.1")
        self._check_conf_file(c, ["myconf_a/0.1", "myconf_b/0.1"])

    def test_update_first(self, servers):
        c = TestClient(servers=servers, light=True)
        c.run("config install-pkg myconf_a/0.1")
        c.run("config install-pkg myconf_b/0.1")

        # update of the first fail
        c.run("config install-pkg myconf_a/[*]", assert_error=True)
        assert "ERROR: Installing these configuration packages will break" in c.out
        assert "use 'conan config install-pkg --force' to force" in c.out
        assert "Use 'conan config clean' first to fully reset your configuration" in c.out
        self._check_conf(c, "myconf_b/0.1")
        self._check_conf_file(c, ["myconf_a/0.1", "myconf_b/0.1"])

        # Now force the first
        c.run("config install-pkg myconf_a/[*] --force")
        assert "Forcing the installation because --force was defined" in c.out
        self._check_conf(c, "myconf_a/0.2")
        self._check_conf_file(c, ["myconf_b/0.1", "myconf_a/0.2"])

    def test_update_second(self, servers):
        c = TestClient(servers=servers, light=True)
        c.run("config install-pkg myconf_a/0.1")
        c.run("config install-pkg myconf_b/0.1")

        # update of the second works without problem
        c.run("config install-pkg myconf_b/[*]")
        assert "Installing new or updating configuration packages" in c.out
        self._check_conf(c, "myconf_b/0.2")
        self._check_conf_file(c, ["myconf_a/0.1", "myconf_b/0.2"])

    def test_error_cant_use_as_dependency(self):
        c = TestClient(light=True)
        conanfile = GenConanfile("myconf", "0.1").with_package_type("configuration")
        c.save({"myconf/conanfile.py": conanfile,
                "pkg/conanfile.py": GenConanfile("pkg", "0.1").with_requires("myconf/0.1")})
        c.run("create myconf")
        c.run("install pkg", assert_error=True)
        assert "ERROR: Configuration package myconf/0.1 cannot be used as requirement, " \
               "but pkg/0.1 is requiring it" in c.out

    def test_error_cant_use_without_type(self):
        c = TestClient(light=True)
        c.save({"myconf/conanfile.py": GenConanfile("myconf", "0.1")})
        c.run("create myconf")
        c.run("config install-pkg myconf/[*]", assert_error=True)
        assert 'ERROR: myconf/0.1 is not of package_type="configuration"' in c.out

    def test_create_also(self):
        conanfile = textwrap.dedent("""
           from conan import ConanFile
           from conan.tools.files import copy
           class Conf(ConanFile):
               name = "myconf"
               version = "0.1"
               package_type = "configuration"
               exports_sources = "*.conf"
               def package(self):
                   copy(self, "*.conf", src=self.build_folder, dst=self.package_folder)
               """)

        c = TestClient(default_server_user=True, light=True)
        c.save({"conanfile.py": conanfile,
                "global.conf": "user.myteam:myconf=my_myconf/0.1_value"})
        c.run("create .")
        c.run("upload * -r=default -c")
        c.run("remove * -c")

        c.run("config install-pkg myconf/[*]")
        self._check_conf(c, "myconf/0.1")
        self._check_conf_file(c, ["myconf/0.1"])

    def test_lockfile(self, servers):
        """ it should be able to install the config older version using a lockfile
        """
        c = TestClient(servers=servers, light=True)
        c.run("config install-pkg myconf_a/0.1 --lockfile-out=config.lock")
        c.run("config install-pkg myconf_a/[*] --lockfile=config.lock")
        assert ("The requested configurations are identical to the already installed ones, "
                "skipping re-installation") in c.out
        self._check_conf(c, "myconf_a/0.1")
        self._check_conf_file(c, ["myconf_a/0.1"])

        # Without the lockfile, it is free to update
        c.run("config install-pkg myconf_a/[*] --lockfile-out=config.lock")
        assert "Installing new or updating configuration packages"
        self._check_conf(c, "myconf_a/0.2")
        self._check_conf_file(c, ["myconf_a/0.2"])
        result = json.loads(c.load("config.lock"))
        assert "myconf_a/0.2" in result["config_requires"][0]

    def test_install_from_file(self, servers):
        c = TestClient(servers=servers, light=True)
        # To make explicit the format of the conanconfig file
        conanconfig = textwrap.dedent("""\
            packages:
                - myconf_a/0.1
                - myconf_b/0.1
            """)
        c.save({"conanconfig.yml": conanconfig})
        # Admits the default location .
        c.run("config install-pkg")
        self._check_conf(c, "myconf_b/0.1")
        self._check_conf_file(c, ["myconf_a/0.1", "myconf_b/0.1"])

        # install same file again:
        c.run("config install-pkg .")
        assert ("The requested configurations are identical to the already installed ones, "
                "skipping re-installation") in c.out
        self._check_conf(c, "myconf_b/0.1")
        self._check_conf_file(c, ["myconf_a/0.1", "myconf_b/0.1"])

        # Forced re-install
        c.save_home({"global.conf": ""})
        c.run("config install-pkg . --force")
        assert "forcing re-installation because --force" in c.out
        self._check_conf(c, "myconf_b/0.1")
        self._check_conf_file(c, ["myconf_a/0.1", "myconf_b/0.1"])

    def test_install_from_file_with_url(self, servers):
        c = TestClient(servers=servers, light=True)
        c.run("remote remove default")
        server_url = servers["default"].fake_url
        conanconfig = yaml.dump({"packages": ["myconf_a/0.1"], "urls": [server_url]})
        c.save({"conanconfig.yml": conanconfig})
        c.run("config install-pkg .")
        self._check_conf(c, "myconf_a/0.1")
        self._check_conf_file(c, ["myconf_a/0.1"])

    def test_update_with_file(self, servers):
        c = TestClient(servers=servers, light=True)
        conanconfig = yaml.dump({"packages": ["myconf_a/0.1", "myconf_b/0.1"]})
        c.save({"conanconfig.yml": conanconfig})
        c.run("config install-pkg .")

        # Now installing "updates" without disrupting the order
        c.run("config install-pkg myconf_c/0.1")
        assert "Installing new or updating configuration packages"
        self._check_conf(c, "myconf_c/0.1")
        self._check_conf_file(c, ["myconf_a/0.1", "myconf_b/0.1", "myconf_c/0.1"])

        # Now installing "updates" without disrupting the order
        conanconfig = textwrap.dedent("""\
            packages:
                - myconf_a/0.1
                - myconf_b/0.1
                - myconf_c/[*]
            """)
        c.save({"conanconfig.yml": conanconfig})
        c.run("config install-pkg .")
        assert "Installing new or updating configuration packages" in c.out
        self._check_conf(c, "myconf_c/0.2")
        self._check_conf_file(c, ["myconf_a/0.1", "myconf_b/0.1", "myconf_c/0.2"])

        # This is also an update of the existing ones, keeping the order
        conanconfig = textwrap.dedent("""\
            packages:
                - myconf_a/[*]
                - myconf_b/0.1
                - myconf_c/[*]
            """)
        c.save({"conanconfig.yml": conanconfig})
        c.run("config install-pkg .")
        assert "Installing new or updating configuration packages" in c.out
        self._check_conf(c, "myconf_c/0.2")
        self._check_conf_file(c, ["myconf_a/0.2", "myconf_b/0.1", "myconf_c/0.2"])

    def test_failed_update_force(self, servers):
        c = TestClient(servers=servers, light=True)
        conanconfig = yaml.dump({"packages": ["myconf_a/0.1", "myconf_b/0.1"]})
        c.save({"conanconfig.yml": conanconfig})
        c.run("config install-pkg .")

        # Now installing "updates" without disrupting the order
        conanconfig = textwrap.dedent("""\
            packages:
                - myconf_a/0.2
            """)
        c.save({"conanconfig.yml": conanconfig})
        c.run("config install-pkg .", assert_error=True)
        assert "use 'conan config install-pkg --force' to force" in c.out
        assert "Use 'conan config clean' first to fully reset" in c.out
        self._check_conf(c, "myconf_b/0.1")
        self._check_conf_file(c, ["myconf_a/0.1", "myconf_b/0.1"])

        c.run("config install-pkg . --force")
        assert "Forcing the installation because --force was defined" in c.out
        self._check_conf(c, "myconf_a/0.2")
        self._check_conf_file(c, ["myconf_b/0.1", "myconf_a/0.2"])

    def test_install_from_file_with_lockfile(self, servers):
        # it should stay in version 0.1
        c = TestClient(servers=servers)
        c.run("config install-pkg myconf_a/0.1 --lockfile-out=conan.lock")
        conanconfig = yaml.dump({"packages": ["myconf_a/[>=0.1 <1.0]"]})
        c.save({"conanconfig.yml": conanconfig})
        c.run("config install-pkg .")
        path = HomePaths(c.cache_folder).config_version_path
        # First are the newest installed
        configs = json.loads(load(path))["config_version"]
        configs = [str(RecipeReference.loads(r)) for r in configs]
        assert configs == ["myconf_a/0.1"]


class TestConfigInstallPkgSettings:
    conanfile = textwrap.dedent("""
        import os
        from conan import ConanFile
        from conan.tools.files import copy
        class Conf(ConanFile):
            name = "myconf"
            version = "0.1"
            settings = "os"
            package_type = "configuration"
            def package(self):
                f = "win" if self.settings.os == "Windows" else "nix"
                copy(self, "*.conf", src=os.path.join(self.build_folder, f), dst=self.package_folder)
            """)

    @pytest.fixture()
    def client(self):
        c = TestClient(default_server_user=True)
        c.save({"conanfile.py": self.conanfile,
                "win/global.conf": "user.myteam:myconf=mywinvalue",
                "nix/global.conf": "user.myteam:myconf=mynixvalue",
                })
        c.run("export-pkg . -s os=Windows")
        c.run("export-pkg . -s os=Linux")
        c.run("upload * -r=default -c")
        c.run("remove * -c")
        return c

    @pytest.mark.parametrize("default_profile", [False, True])
    def test_config_install_from_pkg(self, client, default_profile):
        # Now install it
        c = client
        if not default_profile:
            os.remove(os.path.join(c.cache_folder, "profiles", "default"))
        c.run("config install-pkg myconf/[*] -s os=Windows")
        assert "myconf/0.1: Downloaded package revision" in c.out
        assert "Copying file global.conf" in c.out
        c.run("config show *")
        assert "user.myteam:myconf: mywinvalue" in c.out

        c.run("config install-pkg myconf/[*] -s os=Linux --force")
        assert "myconf/0.1: Downloaded package revision" in c.out
        assert "Copying file global.conf" in c.out
        c.run("config show *")
        assert "user.myteam:myconf: mynixvalue" in c.out

    def test_error_no_settings_defined(self, client):
        c = client
        os.remove(os.path.join(c.cache_folder, "profiles", "default"))
        c.run("config install-pkg myconf/[*]", assert_error=True)
        assert "There are invalid packages:" in c.out
        assert "myconf/0.1: Invalid: 'settings.os' value not defined" in c.out

    def test_config_install_from_pkg_profile(self, client):
        # Now install it
        c = client
        c.save({"win.profile": "[settings]\nos=Windows",
                "nix.profile": "[settings]\nos=Linux"})
        c.run("config install-pkg myconf/[*] -pr=win.profile")
        assert "myconf/0.1: Downloaded package revision" in c.out
        assert "Copying file global.conf" in c.out
        c.run("config show *")
        assert "user.myteam:myconf: mywinvalue" in c.out

        c.run("config install-pkg myconf/[*] -pr=nix.profile --force")
        assert "myconf/0.1: Downloaded package revision" in c.out
        assert "Copying file global.conf" in c.out
        c.run("config show *")
        assert "user.myteam:myconf: mynixvalue" in c.out

    def test_config_install_from_pkg_profile_default(self, client):
        # Now install it
        c = client
        c.save_home({"profiles/default": "[settings]\nos=Windows"})
        c.run("config install-pkg myconf/[*]")
        assert "myconf/0.1: Downloaded package revision" in c.out
        assert "Copying file global.conf" in c.out
        c.run("config show *")
        assert "user.myteam:myconf: mywinvalue" in c.out

        c.save_home({"profiles/default": "[settings]\nos=Linux"})
        c.run("config install-pkg myconf/[*] --force")
        assert "myconf/0.1: Downloaded package revision" in c.out
        assert "Copying file global.conf" in c.out
        c.run("config show *")
        assert "user.myteam:myconf: mynixvalue" in c.out


class TestConfigInstallPkgOptions:
    conanfile = textwrap.dedent("""
        import os
        from conan import ConanFile
        from conan.tools.files import copy
        class Conf(ConanFile):
            name = "myconf"
            version = "0.1"
            options = {"project": ["project1", "project2"]}
            default_options = {"project": "project1"}
            package_type = "configuration"
            def package(self):
                copy(self, "*.conf", src=os.path.join(self.build_folder, str(self.options.project)),
                     dst=self.package_folder)
            """)

    @pytest.fixture()
    def client(self):
        c = TestClient(default_server_user=True)
        c.save({"conanfile.py": self.conanfile,
                "project1/global.conf": "user.myteam:myconf=my1value",
                "project2/global.conf": "user.myteam:myconf=my2value",
                })
        c.run("export-pkg .")
        c.run("export-pkg . -o project=project2")
        c.run("upload * -r=default -c")
        c.run("remove * -c")
        return c

    @pytest.mark.parametrize("default_profile", [False, True])
    def test_config_install_from_pkg(self, client, default_profile):
        # Now install it
        c = client
        if not default_profile:
            os.remove(os.path.join(c.cache_folder, "profiles", "default"))
        c.run("config install-pkg myconf/[*] -o &:project=project1")
        assert "myconf/0.1: Downloaded package revision" in c.out
        assert "Copying file global.conf" in c.out
        c.run("config show *")
        assert "user.myteam:myconf: my1value" in c.out

        c.run("config install-pkg myconf/[*] -o &:project=project2 --force")
        assert "myconf/0.1: Downloaded package revision" in c.out
        assert "Copying file global.conf" in c.out
        c.run("config show *")
        assert "user.myteam:myconf: my2value" in c.out

    def test_no_option_defined(self, client):
        c = client
        os.remove(os.path.join(c.cache_folder, "profiles", "default"))
        c.run("config install-pkg myconf/[*]")
        assert "Copying file global.conf" in c.out
        c.run("config show *")
        assert "user.myteam:myconf: my1value" in c.out

    def test_config_install_from_pkg_profile(self, client):
        # Now install it
        c = client
        c.save({"win.profile": "[options]\n&:project=project1",
                "nix.profile": "[options]\n&:project=project2"})
        c.run("config install-pkg myconf/[*] -pr=win.profile")
        assert "myconf/0.1: Downloaded package revision" in c.out
        assert "Copying file global.conf" in c.out
        c.run("config show *")
        assert "user.myteam:myconf: my1value" in c.out

        c.run("config install-pkg myconf/[*] -pr=nix.profile --force")
        assert "myconf/0.1: Downloaded package revision" in c.out
        assert "Copying file global.conf" in c.out
        c.run("config show *")
        assert "user.myteam:myconf: my2value" in c.out

    def test_config_install_from_pkg_profile_default(self, client):
        # Now install it
        c = client
        c.save_home({"profiles/default": "[options]\n&:project=project1"})
        c.run("config install-pkg myconf/[*]")
        assert "myconf/0.1: Downloaded package revision" in c.out
        assert "Copying file global.conf" in c.out
        c.run("config show *")
        assert "user.myteam:myconf: my1value" in c.out

        c.save_home({"profiles/default": "[options]\n&:project=project2"})
        c.run("config install-pkg myconf/[*] --force")
        assert "myconf/0.1: Downloaded package revision" in c.out
        assert "Copying file global.conf" in c.out
        c.run("config show *")
        assert "user.myteam:myconf: my2value" in c.out



import os
import platform
import textwrap

import pytest

from conan.internal.paths import CONANFILE
from conan.test.utils.tools import TestClient, GenConanfile


class TestBuildRequires:

    @pytest.fixture()
    def client(self):
        c = TestClient()
        tool_conanfile = textwrap.dedent("""
            import os
            from conan import ConanFile
            from conan.tools.files import copy

            class tool(ConanFile):
                name = "tool"
                version = "0.1"
                exports_sources = "mytool*"

                def package(self):
                    copy(self, "mytool*", self.source_folder, self.package_folder)

                def package_info(self):
                    self.buildenv_info.append_path("PATH", self.package_folder)
            """)
        name = "mytool.bat" if platform.system() == "Windows" else "mytool"
        c.save({CONANFILE: tool_conanfile,
                name: "echo Hello World!"}, clean_first=True)
        os.chmod(os.path.join(c.current_folder, name), 0o777)
        c.run("export . --user=lasote --channel=stable")

        lib_conanfile = textwrap.dedent("""
            from conan import ConanFile

            class mylib(ConanFile):
                name = "mylib"
                version = "0.1"

                def build(self):
                    self.run("mytool")
            """)
        profile = """
            [tool_requires]
            tool/0.1@lasote/stable
            nonexistingpattern*: sometool/1.2@user/channel
            """

        profile2 = """
            [tool_requires]
            tool/0.1@lasote/stable
            nonexistingpattern*: sometool/1.2@user/channel
            """

        test_conanfile = textwrap.dedent("""
           from conan import ConanFile

           class Testmylib(ConanFile):

               def build(self):
                   self.run("mytool")
               def test(self):
                   pass
           """)
        c.save({CONANFILE: lib_conanfile,
                "test_package/conanfile.py": test_conanfile,
                "profile.txt": profile,
                "profile2.txt": profile2}, clean_first=True)
        return c

    def test_profile_requires(self, client):
        """
        cli -(tool-requires)-> tool/0.1
          \\--(requires)->mylib/0.1 -(tool_requires)->tool/0.1 (skipped)
        """
        client.run("export . --user=lasote --channel=stable")
        client.run("install --requires=mylib/0.1@lasote/stable --profile ./profile.txt --build missing")
        assert "Hello World!" in client.out

        client.run("install --requires=mylib/0.1@lasote/stable --profile ./profile2.txt --build='*'")
        assert "Hello World!" in client.out

    def test_profile_open_requires(self, client):
        client.run("build . --profile ./profile.txt --build missing")
        assert "Hello World!" in client.out

    def test_build_mode_requires(self, client):
        client.run("install . --profile ./profile.txt", assert_error=True)
        assert "ERROR: Missing prebuilt package for 'tool/0.1@lasote/stable'" in client.out
        client.run("install . --profile ./profile.txt --build=Pythontool", assert_error=True)
        assert "ERROR: Missing prebuilt package for 'tool/0.1@lasote/stable'" in client.out
        client.run("install . --profile ./profile.txt --build=tool/0.1*")
        assert "tool/0.1@lasote/stable: Created package" in client.out

        # now remove packages, ensure --build=missing also creates them
        client.run('remove "*:*" -c')
        client.run("install . --profile ./profile.txt --build=missing")
        assert "tool/0.1@lasote/stable: Created package" in client.out

    def test_profile_test_requires(self, client):
        client.run("create . --profile ./profile.txt --build missing")
        assert 2 == str(client.out).splitlines().count("Hello World!")

    def test_consumer_patterns(self, client):
        profile_patterns = """
            [tool_requires]
            &: tool/0.1@lasote/stable
            nonexistingpattern*: sometool/1.2@user/channel
            """
        client.save({CONANFILE: GenConanfile("mylib", "0.1"),
                     "profile.txt": profile_patterns})
        client.run("create . --profile=./profile.txt --build=missing")
        assert 1 == str(client.out).splitlines().count("Hello World!")

    def test_build_requires_options(self):
        client = TestClient()
        client.save({CONANFILE: GenConanfile("mytool", "0.1")})
        client.run("export . --user=lasote --channel=stable")

        conanfile = textwrap.dedent("""
            from conan import ConanFile

            class mylib(ConanFile):
                name = "mylib"
                version = "0.1"
                build_requires = "mytool/0.1@lasote/stable"
                options = {"coverage": [True, False]}
                def build(self):
                    self.output.info("Coverage %s" % self.options.coverage)
            """)
        client.save({CONANFILE: conanfile}, clean_first=True)
        client.run("build . -o mylib*:coverage=True --build missing")
        client.assert_listed_require({"mytool/0.1@lasote/stable": "Cache"}, build=True)
        assert "conanfile.py (mylib/0.1): Coverage True" in client.out

        client.save({CONANFILE: conanfile}, clean_first=True)
        client.run("build . -o coverage=True")
        client.assert_listed_require({"mytool/0.1@lasote/stable": "Cache"}, build=True)
        assert "mytool/0.1@lasote/stable: Already installed!" in client.out
        assert "conanfile.py (mylib/0.1): Coverage True" in client.out


def test_consumer_patterns_loop_error():
    client = TestClient()

    profile_patterns = textwrap.dedent("""
        include(default)
        [tool_requires]
        tool1/1.0
        tool2/1.0
        """)
    client.save({"tool1/conanfile.py": GenConanfile(),
                 "tool2/conanfile.py": GenConanfile().with_build_requires("tool1/1.0"),
                 "consumer/conanfile.py": GenConanfile(),
                 "profile.txt": profile_patterns})

    client.run("export tool1 --name=tool1 --version=1.0")
    client.run("export tool2 --name=tool2 --version=1.0")
    client.run("install consumer --build=missing -pr:b=profile.txt -pr:h=profile.txt",
               assert_error=True)
    assert "There is a cycle/loop in the graph" in client.out

    # we can fix it with the negation
    profile_patterns = textwrap.dedent("""
        include(default)
        [tool_requires]
        tool1/1.0
        !tool1*:tool2/1.0
        """)
    client.save({"profile.txt": profile_patterns})

    client.run("install consumer --build=missing -pr:b=profile.txt -pr:h=profile.txt")
    assert "tool1/1.0: Created package" in client.out
    assert "tool2/1.0: Created package" in client.out


def test_tool_requires_revision_profile():
    # We shoul be able to explicitly [tool_require] a recipe revision in the profile
    c = TestClient()
    build_profile = textwrap.dedent("""\
        [settings]
        os=Linux
        [tool_requires]
        *:tool/0.1#2d65f1b4af1ce59028f96adbfe7ed5a2
        """)
    c.save({"tool/conanfile.py": GenConanfile("tool", "0.1"),
            "cmake/conanfile.py": GenConanfile("cmake", "0.1"),
            "app/conanfile.py": GenConanfile("app", "0.1").with_tool_requires("cmake/0.1"),
            "build_profile": build_profile})
    c.run("export tool")
    rev1 = c.exported_recipe_revision()
    assert rev1 == "2d65f1b4af1ce59028f96adbfe7ed5a2"
    # Create a new tool revision to proof that we can still require the old one
    c.save({"tool/conanfile.py": GenConanfile("tool", "0.1").with_class_attribute("myvar=42")})
    c.run("export tool")
    rev2 = c.exported_recipe_revision()
    assert rev2 != rev1
    c.run("export cmake")
    c.run("graph info app -pr:b=build_profile --build=*")
    assert f"tool/0.1#{rev1}" in c.out
    assert rev2 not in c.out


def test_tool_requires_version_range_loop():
    # https://github.com/conan-io/conan/issues/17930
    c = TestClient(light=True)
    build_profile = textwrap.dedent("""\
        [settings]
        os=Linux
        [tool_requires]
        tool/[>=1.0 <2]
        """)
    c.save({"tool/conanfile.py": GenConanfile("tool", "1.1"),
            "app/conanfile.py": GenConanfile("app", "0.1").with_tool_requires("tool/1.1"),
            "build_profile": build_profile})
    c.run("create tool")
    c.run("install app -pr:b=build_profile")
    assert "tool/1.1" in c.out  # It is skipped



import json
import os

from conan.api.conan_api import ConanAPI
from conan.api.model import RecipeReference
from conan.api.output import ConanOutput, cli_out_write
from conan.api.subapi.workspace import WorkspaceAPI
from conan.cli import make_abs_path
from conan.cli.args import add_reference_args, add_common_install_arguments, add_lockfile_args
from conan.cli.command import conan_command, conan_subcommand
from conan.cli.commands.list import print_serial
from conan.cli.formatters.graph import format_graph_json
from conan.cli.printers import print_profiles
from conan.cli.printers.graph import print_graph_packages, print_graph_basic
from conan.errors import ConanException
from conan.internal.graph.install_graph import ProfileArgs


@conan_subcommand(formatters={"text": cli_out_write})
def workspace_root(conan_api: ConanAPI, parser, subparser, *args):  # noqa
    """
    Return the folder containing the conanws.py/conanws.yml workspace file
    """
    parser.parse_args(*args)
    ws = conan_api.workspace
    return ws.folder()


@conan_subcommand()
def workspace_open(conan_api: ConanAPI, parser, subparser, *args):
    """
    Open specific references
    """
    subparser.add_argument("reference",
                           help="Open this package source repository")
    group = subparser.add_mutually_exclusive_group()
    group.add_argument("-r", "--remote", action="append", default=None,
                       help='Look in the specified remote or remotes server')
    group.add_argument("-nr", "--no-remote", action="store_true",
                       help='Do not use remote, resolve exclusively in the cache')
    args = parser.parse_args(*args)
    remotes = conan_api.remotes.list(args.remote) if not args.no_remote else []
    cwd = os.getcwd()
    conan_api.workspace.open(args.reference, remotes=remotes, cwd=cwd)


@conan_subcommand()
def workspace_add(conan_api: ConanAPI, parser, subparser, *args):
    """
    Add packages to current workspace
    """
    subparser.add_argument('path',  nargs="?",
                           help='Path to the package folder in the user workspace')
    add_reference_args(subparser)
    subparser.add_argument("--ref", help="Open and add this reference")
    subparser.add_argument("-of", "--output-folder",
                           help='The root output folder for generated and build files')
    group = subparser.add_mutually_exclusive_group()
    group.add_argument("-r", "--remote", action="append", default=None,
                       help='Look in the specified remote or remotes server')
    group.add_argument("-nr", "--no-remote", action="store_true",
                       help='Do not use remote, resolve exclusively in the cache')
    args = parser.parse_args(*args)
    if args.path and args.ref:
        raise ConanException("Do not use both 'path' and '--ref' argument")
    remotes = conan_api.remotes.list(args.remote) if not args.no_remote else []
    cwd = os.getcwd()
    path = args.path
    if args.ref:
        # TODO: Use path here to open in this path
        path = conan_api.workspace.open(args.ref, remotes, cwd=cwd)
    ref = conan_api.workspace.add(path,
                                  args.name, args.version, args.user, args.channel,
                                  cwd, args.output_folder, remotes=remotes)
    ConanOutput().success("Reference '{}' added to workspace".format(ref))


@conan_subcommand()
def workspace_complete(conan_api: ConanAPI, parser, subparser, *args):
    """
    Complete the workspace, opening or adding intermediate packages to it that have
    requirements to other packages in the workspace.
    """
    add_common_install_arguments(subparser)
    add_lockfile_args(subparser)
    args = parser.parse_args(*args)
    remotes = conan_api.remotes.list(args.remote) if not args.no_remote else []
    overrides = eval(args.lockfile_overrides) if args.lockfile_overrides else None
    # The lockfile by default if not defined will be read from the root workspace folder
    ws_folder = conan_api.workspace.folder()
    lockfile = conan_api.lockfile.get_lockfile(lockfile=args.lockfile, conanfile_path=ws_folder,
                                               cwd=None, partial=args.lockfile_partial,
                                               overrides=overrides)
    profile_host, profile_build = conan_api.profiles.get_profiles_from_args(args)
    print_profiles(profile_host, profile_build)

    ConanOutput().box("Workspace completing intermediate packages")
    conan_api.workspace.complete(profile_host, profile_build, lockfile, remotes, update=args.update)


@conan_subcommand()
def workspace_remove(conan_api: ConanAPI, parser, subparser, *args):
    """
    Remove packages from the current workspace
    """
    subparser.add_argument('path',
                           help='Path to the package folder in the user workspace')
    args = parser.parse_args(*args)
    removed = conan_api.workspace.remove(make_abs_path(args.path))
    ConanOutput().info(f"Removed from workspace: {removed}")


def _print_json(data):
    results = data["info"]
    myjson = json.dumps(results, indent=4)
    cli_out_write(myjson)


def _print_workspace_info(data):
    ret = []
    for package_info in data["info"]["packages"]:
        ret.append(f"- path: {package_info['path']}")
        ref = package_info.get('ref')
        if ref:
            ret.append(f"  ref: {ref}")
    data["info"]["packages"] = ret or "(empty)"
    print_serial(data["info"])


@conan_subcommand(formatters={"text": _print_workspace_info, "json": _print_json})
def workspace_info(conan_api: ConanAPI, parser, subparser, *args):  # noqa
    """
    Display info for current workspace
    """
    parser.parse_args(*args)
    return {"info": conan_api.workspace.info()}


@conan_subcommand()
def workspace_build(conan_api: ConanAPI, parser, subparser, *args):
    """
    Call "conan build" for packages in the workspace, in the right order
    """
    _install_build(conan_api, parser, subparser, True, *args)


@conan_subcommand()
def workspace_install(conan_api: ConanAPI, parser, subparser, *args):
    """
    Call "conan install" for packages in the workspace, in the right order
    """
    _install_build(conan_api, parser, subparser, False, *args)


def _install_build(conan_api: ConanAPI, parser, subparser, build, *args):
    subparser.add_argument("--pkg", action="append", help='Define specific packages')
    add_common_install_arguments(subparser)
    add_lockfile_args(subparser)
    args = parser.parse_args(*args)
    # Basic collaborators: remotes, lockfile, profiles
    remotes = conan_api.remotes.list(args.remote) if not args.no_remote else []
    overrides = eval(args.lockfile_overrides) if args.lockfile_overrides else None
    # The lockfile by default if not defined will be read from the root workspace folder
    ws_folder = conan_api.workspace.folder()
    lockfile = conan_api.lockfile.get_lockfile(lockfile=args.lockfile, conanfile_path=ws_folder,
                                               cwd=None, partial=args.lockfile_partial,
                                               overrides=overrides)
    profile_host, profile_build = conan_api.profiles.get_profiles_from_args(args)
    print_profiles(profile_host, profile_build)

    buildmode = args.build
    if build and (not buildmode or "editable" not in buildmode):
        ConanOutput().info("Adding '--build=editable' as build mode")
        buildmode = buildmode or []
        buildmode.append("editable")

    all_editables = conan_api.workspace.packages()
    packages = conan_api.workspace.select_packages(args.pkg)
    ConanOutput().box("Workspace computing the build order")
    install_order = conan_api.workspace.build_order(packages, profile_host, profile_build, buildmode,
                                                    lockfile, remotes, args, update=args.update)

    msg = "build" if build else "install"
    ConanOutput().box(f"Workspace {msg}ing each package")
    order = install_order.install_build_order()

    profile_args = ProfileArgs.from_args(args)
    for level in order["order"]:
        for elem in level:
            ref = RecipeReference.loads(elem["ref"])
            for package_level in elem["packages"]:
                for package in package_level:
                    ws_pkg = all_editables.get(ref)
                    is_editable = package["binary"] in ("Editable", "EditableBuild")
                    if ws_pkg is None:
                        if is_editable or package["binary"] == "Build": # Build external to Workspace
                            cmd = f'install {package["build_args"]} {profile_args}'
                            ConanOutput().box(f"Workspace building external {ref}")
                            ConanOutput().info(f"Command: {cmd}\n")
                            conan_api.command.run(cmd)
                    else:
                        path = ws_pkg["path"]
                        output_folder = ws_pkg.get("output_folder")
                        build_arg = "--build-require" if package["context"] == "build" else ""
                        ref_args = " ".join(f"--{k}={getattr(ref, k)}"
                                            for k in ("name", "version", "user", "channel")
                                            if getattr(ref, k, None))
                        of_arg = f'-of="{output_folder}"' if output_folder else ""
                        # TODO: Missing --lockfile-overrides arg here
                        command = "build" if build else "install"
                        cmd = f'{command} "{path}" {profile_args} {build_arg} {ref_args} {of_arg}'
                        ConanOutput().box(f"Workspace {command}: {ref}")
                        ConanOutput().info(f"Command: {cmd}\n")
                        conan_api.command.run(cmd)


@conan_subcommand(formatters={"json": format_graph_json})
def workspace_super_install(conan_api: ConanAPI, parser, subparser, *args):
    """
    Install the workspace as a monolith, installing only external dependencies to the workspace,
    generating a single result (generators, etc) for the whole workspace.
    """
    subparser.add_argument("--pkg", action="append", help='Define specific packages')
    subparser.add_argument("-g", "--generator", action="append", help='Generators to use')
    subparser.add_argument("-of", "--output-folder",
                           help='The root output folder for generated and build files')
    subparser.add_argument("-d", "--deployer", action="append",
                           help="Deploy using the provided deployer to the output folder. "
                                "Built-in deployers: 'full_deploy', 'direct_deploy', "
                                "'runtime_deploy'")
    subparser.add_argument("--deployer-folder",
                           help="Deployer output folder, base build folder by default if not set")
    subparser.add_argument("--deployer-package", action="append",
                           help="Execute the deploy() method of the packages matching "
                                "the provided patterns")
    subparser.add_argument("--envs-generation", default=None, choices=["false"],
                           help="Generation strategy for virtual environment files for the root")
    add_common_install_arguments(subparser)
    add_lockfile_args(subparser)
    args = parser.parse_args(*args)
    # Basic collaborators: remotes, lockfile, profiles
    remotes = conan_api.remotes.list(args.remote) if not args.no_remote else []
    overrides = eval(args.lockfile_overrides) if args.lockfile_overrides else None
    # The lockfile by default if not defined will be read from the root workspace folder
    ws_folder = conan_api.workspace.folder()
    lockfile = conan_api.lockfile.get_lockfile(lockfile=args.lockfile, conanfile_path=ws_folder,
                                               cwd=None,
                                               partial=args.lockfile_partial, overrides=overrides)
    profile_host, profile_build = conan_api.profiles.get_profiles_from_args(args)
    print_profiles(profile_host, profile_build)

    # Build a dependency graph with all editables as requirements
    requires = conan_api.workspace.select_packages(args.pkg)
    deps_graph = conan_api.graph.load_graph_requires(requires, [],
                                                     profile_host, profile_build, lockfile,
                                                     remotes, update=args.update)
    deps_graph.report_graph_error()
    print_graph_basic(deps_graph)

    # Collapsing the graph
    ws_graph = conan_api.workspace.super_build_graph(deps_graph, profile_host, profile_build)
    ConanOutput().subtitle("Collapsed graph")
    print_graph_basic(ws_graph)
    conan_api.graph.analyze_binaries(ws_graph, args.build, remotes=remotes, update=args.update,
                                     lockfile=lockfile)
    print_graph_packages(ws_graph)
    conan_api.install.install_binaries(deps_graph=ws_graph, remotes=remotes)
    ConanOutput().title("Finalizing install (deploy, generators)")
    output_folder = make_abs_path(args.output_folder) if args.output_folder else None
    conan_api.install.install_consumer(ws_graph, args.generator, ws_folder, output_folder,
                                       deploy=args.deployer, deploy_package=args.deployer_package,
                                       deploy_folder=args.deployer_folder,
                                       envs_generation=args.envs_generation)
    ConanOutput().success("Install finished successfully")

    return {"graph": ws_graph,
            "conan_api": conan_api}


@conan_subcommand()
def workspace_clean(conan_api: ConanAPI, parser, subparser, *args):  # noqa
    """
    Clean the temporary build folders when possible
    """
    parser.parse_args(*args)
    conan_api.workspace.clean()


@conan_subcommand()
def workspace_init(conan_api: ConanAPI, parser, subparser, *args):
    """
    Clean the temporary build folders when possible
    """
    subparser.add_argument("path", nargs="?", default=os.getcwd(),
                           help="Path to a folder where the workspace will be initialized. "
                                "Defaults to the current directory")
    args = parser.parse_args(*args)
    conan_api.workspace.init(args.path)


@conan_subcommand()
def workspace_create(conan_api: ConanAPI, parser, subparser, *args):
    """
    Call "conan create" for packages in the workspace, in the correct order.
    Packages will be created in the Conan cache, not locally
    """
    subparser.add_argument("--pkg", action="append", help='Define specific packages')
    add_common_install_arguments(subparser)
    add_lockfile_args(subparser)
    args = parser.parse_args(*args)
    # Basic collaborators: remotes, lockfile, profiles
    remotes = conan_api.remotes.list(args.remote) if not args.no_remote else []
    overrides = eval(args.lockfile_overrides) if args.lockfile_overrides else None
    # The lockfile by default if not defined will be read from the root workspace folder
    ws_folder = conan_api.workspace.folder()
    lockfile = conan_api.lockfile.get_lockfile(lockfile=args.lockfile, conanfile_path=ws_folder,
                                               cwd=None,
                                               partial=args.lockfile_partial, overrides=overrides)
    profile_host, profile_build = conan_api.profiles.get_profiles_from_args(args)
    print_profiles(profile_host, profile_build)

    build_mode = args.build if args.build else []
    ConanOutput().box("Exporting workspace recipes to Conan cache")
    exported_refs = conan_api.workspace.export()
    build_mode.extend(f"missing:{r}" for r in exported_refs)

    all_packages = conan_api.workspace.packages()
    packages = conan_api.workspace.select_packages(args.pkg)

    # If we don't disable the workspace, then, the packages are not created in the Conan cache,
    # but locally in the user folders, are they are intercepted as editables
    conan_api.workspace.enable(False)

    install_order = conan_api.workspace.build_order(packages, profile_host, profile_build,
                                                    build_mode, lockfile, remotes, args,
                                                    update=args.update)

    ConanOutput().box("Workspace creating each package")
    order = install_order.install_build_order()

    profile_args = ProfileArgs.from_args(args)
    for level in order["order"]:
        for elem in level:
            ref = RecipeReference.loads(elem["ref"])
            for package_level in elem["packages"]:
                for package in package_level:
                    if package["binary"] not in ("Build", "EditableBuild"):
                        ConanOutput().info(f"Skip build for {ref} binary: {package['binary']}")
                        continue

                    if ref not in all_packages:
                        # Build external to Workspace
                        cmd = f'install {package["build_args"]} {profile_args}'
                        ConanOutput().box(f"Workspace building external {ref}")
                        ConanOutput().info(f"Build command: {cmd}\n")
                        conan_api.command.run(cmd)
                    else:  # Package in workspace
                        path = packages[ref]["path"]
                        # TODO: Missing --lockfile-overrides arg here
                        build = "--build-require" if package["context"] == "build" else ""
                        ref_args = " ".join(f"--{k}={getattr(ref, k)}"
                                            for k in ("name", "version", "user", "channel")
                                            if getattr(ref, k, None))
                        cmd = f'create "{path}" {profile_args} {build} {ref_args}'
                        ConanOutput().box(f"Workspace create {ref}")
                        ConanOutput().info(f"Conan create command: {cmd}\n")
                        conan_api.command.run(cmd)


@conan_subcommand()
def workspace_source(conan_api: ConanAPI, parser, subparser, *args):
    """
    Call the source() method of packages in the workspace
    """
    subparser.add_argument("--pkg", action="append", help='Define specific packages')
    args = parser.parse_args(*args)

    remotes = conan_api.remotes.list()  # In case "python_requires" are needed
    packages = conan_api.workspace.select_packages(args.pkg)

    ConanOutput().box("Workspace getting sources")
    for pkg, info in packages.items():
        conan_api.local.source(info["path"], name=pkg.name, version=str(pkg.version),
                               user=pkg.user, channel=pkg.channel, remotes=remotes)


@conan_command(group="Consumer")
def workspace(conan_api, parser, *args):  # noqa
    """
    Manage Conan workspaces (group of packages in editable mode)
    """
    if (WorkspaceAPI.TEST_ENABLED or os.getenv("CONAN_WORKSPACE_ENABLE")) != "will_break_next":
        raise ConanException("Workspace command disabled without CONAN_WORKSPACE_ENABLE env var,"
                             "please read the docs about this 'incubating' feature")
